from typing import List

import numpy
from PIL import Image
from torchvision import transforms as T
from torchvision.transforms import functional as F

__all__ = ["apply_op"]


def gaussian_blur(
    image: Image.Image, sigma: float = 0.0, truncate: float = 3.5
) -> Image.Image:
    """Apply gaussian blur to an image.

    Args:
        image: Input image.
        sigma: Sigma, used to calculate kernel_size. Defaults to 0.0.
        truncate: Truncate kernel. Defaults to 3.5.

    Returns:
        Blurred image.
    """
    # Check image.
    if not isinstance(image, Image.Image):
        raise TypeError("Expected a PIL image not {}".format(type(image)))
    # Define kernel size.
    kernel_size = round(float(sigma) * truncate)
    kernel_size = max(3, kernel_size // 2 * 2 + 1)
    if sigma > 0:
        image = F.gaussian_blur(
            image, kernel_size=(kernel_size, kernel_size), sigma=sigma
        )
    return image


def adjust_channel(image: Image.Image, factor: float, channel: int) -> Image.Image:
    """Adjust image channel values by a factor.

    Args:
        image: Input image.
        factor: Multiplier factor.
        channel: Channel number.

    Returns:
        Adjusted image.
    """
    if not isinstance(image, Image.Image):
        raise TypeError("Expected a PIL image not {}".format(type(image)))
    # Convert to array.
    arr = numpy.array(image).astype(numpy.float32) / 255
    # Multiply
    if arr.ndim > 2:
        arr[..., channel] *= factor
    else:
        arr *= factor
    # Clip values and convert back to uint8.
    image = (numpy.clip(arr, 0.0, 1.0) * 255).astype(numpy.uint8)
    return Image.fromarray(image)


def apply_op(
    image: Image.Image,
    name: str,
    magnitude: float,
    interpolation: T.InterpolationMode = T.InterpolationMode.NEAREST,
    fill: List[int] = [128, 128, 128],
) -> Image.Image:
    """Apply operation to an image with magnitude.

    Args:
        image: Input image.
        name: Operation name.
        magnitude: Magnitude for the operation
        interpolation: Interpolation mode. Defaults to NEAREST.
        fill: Fill value. Defaults to [128, 128, 128].

    Returns:
        Transformed image.
    """
    # Check image.
    if not isinstance(image, Image.Image):
        raise TypeError("Expected a PIL image not {}".format(type(image)))
    if name.lower() == "shearx":
        image = F.affine(
            image,
            angle=0.0,
            translate=[0, 0],
            scale=1.0,
            shear=[magnitude, 0.0],
            interpolation=interpolation,
            fill=fill,
        )
    elif name.lower() == "sheary":
        image = F.affine(
            image,
            angle=0.0,
            translate=[0, 0],
            scale=1.0,
            shear=[0.0, magnitude],
            interpolation=interpolation,
            fill=fill,
        )
    elif name.lower() == "translatex":
        image = F.affine(
            image,
            angle=0.0,
            translate=[int(magnitude), 0],
            scale=1.0,
            interpolation=interpolation,
            shear=[0.0, 0.0],
            fill=fill,
        )
    elif name.lower() == "translatey":
        image = F.affine(
            image,
            angle=0.0,
            translate=[0, int(magnitude)],
            scale=1.0,
            interpolation=interpolation,
            shear=[0.0, 0.0],
            fill=fill,
        )
    elif name.lower() == "rotate":
        image = F.rotate(image, int(magnitude), interpolation=interpolation, fill=fill)
    elif name.lower() == "brightness":
        image = F.adjust_brightness(image, magnitude)
    elif name.lower() == "saturation":
        image = F.adjust_saturation(image, magnitude)
    elif name.lower() == "grayscale":
        image = F.adjust_saturation(image, 0.0)
    elif name.lower() == "hue":
        # HSV conversion causes changes.
        if magnitude != 0:
            image = F.adjust_hue(image, magnitude)
    elif name.lower() == "gamma":
        image = F.adjust_gamma(image, magnitude)
    elif name.lower() == "contrast":
        image = F.adjust_contrast(image, magnitude)
    elif name.lower() == "red":
        image = adjust_channel(image, magnitude, channel=0)
    elif name.lower() == "green":
        image = adjust_channel(image, magnitude, channel=1)
    elif name.lower() == "blue":
        image = adjust_channel(image, magnitude, channel=2)
    elif name.lower() == "sharpness":
        image = F.adjust_sharpness(image, magnitude)
    elif name.lower() == "blur" or name.lower() == "gaussian_blur":
        image = gaussian_blur(image, magnitude)
    elif name.lower() == "posterize":
        image = F.posterize(image, int(magnitude))
    elif name.lower() == "solarize":
        if int(magnitude) <= 255:
            image = F.solarize(image, int(magnitude))
    elif name.lower() == "autocontrast":
        image = F.autocontrast(image)
    elif name.lower() == "equalize":
        image = F.equalize(image)
    elif name.lower() == "invert":
        image = F.invert(image)
    elif name.lower() == "identity":
        pass
    else:
        raise ValueError(f"Operation '{name}' not recognized.")
    return image



import os
from typing import Callable, List, Tuple

from PIL import Image

from ._multiprocess import multiprocess_loop
from ._progress import progress_bar

__all__ = ["shift_dataset"]


def process_image(path: str, output_dir: str, function: Callable):
    """Helper function to load, shift and save an image."""
    try:
        image = Image.open(path)
        shifted = function(image)
        shifted.save(os.path.join(output_dir, os.path.basename(path)))
    except Exception as e:
        return path, e
    return path, None


def shift_dataset(
    paths: List[str], output_dir: str, function: Callable, num_workers: int = 1
) -> List[Tuple[str, Exception]]:
    """Shift the data distribution of dataset to a known direction with a function.

    Args:
        paths: Paths to images in the dataset.
        output_dir: Output directory for the new dataset images.
        function: Function used to shift the data distribution.
        num_processes: Number worker processes.

    Returns:
        List of paths, which could not be processed with the associated Expection.
    """
    # Check output_dir.
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    elif not os.path.isdir(output_dir):
        raise NotADirectoryError("{} exists but isn't a directory.")
    # Shift images.
    exceptions = []
    for logs, (path, exception) in progress_bar(
        multiprocess_loop(
            func=process_image,
            iterable=paths,
            output_dir=output_dir,
            function=function,
            num_workers=num_workers,
        ),
        total=len(paths),
        desc="Processing images",
        log_values=True,
    ):
        if exception is not None:
            exceptions.append((path, exception))
            logs["failures"] = len(exceptions)
    return exceptions



import os
import shutil
from functools import partial

import torchvision.transforms.functional as F
from strong_augment import shift_dataset

IMAGE_PATH = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "strong_augment", "image.jpeg"
)
TMP_DIR = os.path.join(os.path.dirname(__file__), "tmp")


def test_distribution_shift():
    if os.path.exists(TMP_DIR):
        shutil.rmtree(TMP_DIR)
    # Shift function.
    output = shift_dataset(
        paths=[IMAGE_PATH] * 100,
        output_dir=TMP_DIR,
        function=partial(F.adjust_gamma, gamma=0.2),
        num_workers=1,
    )
    assert output == []
    if os.path.exists(TMP_DIR):
        shutil.rmtree(TMP_DIR)



import random
from typing import Callable, Dict, List, Tuple, Union

import numpy
from PIL import Image
from torchvision.transforms import InterpolationMode

from ._operations import apply_op

AFFINE_TRANSFORMS = {
    "shearx",
    "sheary",
    "translatex",
    "translatey",
    "perspective",
    "rotate",
}
AUGMENTATION_SPACE = {
    "identity": (True, False),
    "red": (0.0, 2.0),
    "green": (0.0, 2.0),
    "blue": (0.0, 2.0),
    "hue": (-0.5, 0.5),
    "saturation": (0.0, 2.0),
    "brightness": (0.1, 2.0),
    "contrast": (0.1, 2.0),
    "gamma": (0.1, 2.0),
    "sharpness": (1.0, 4.0),
    "blur": (0.0, 2.0),
    "solarize": (0, 256),
    "posterize": (1, 8),
    "autocontrast": (True, False),
    "equalize": (True, False),
    "grayscale": (True, False),
    "shearx": (-45.0, 45.0),
    "sheary": (-45.0, 45.0),
    "translatex": (-32.0, 32.0),
    "translatey": (-32.0, 32.0),
    "rotate": (-135.0, 135.0),
}

__all__ = ["StrongAugment", "AUGMENTATION_SPACE", "augmentation_collage"]


class StrongAugment:
    def __init__(
        self,
        p: float = 0.4,
        min_ops: int = 2,
        max_ops: int = 5,
        max_affine: int = 1,
        interpolation: InterpolationMode = InterpolationMode.NEAREST,
        augmentation_space: Dict[
            str, List[Union[int, float, bool]]
        ] = AUGMENTATION_SPACE,
        fill: List[int] = [128, 128, 128],
    ):
        """Strong automatic augmentation.

        Args:
            p: Probability consecutive ops after `min_ops`. Defaults  to 0.4.
            min_ops: Minimum number of operations. Defaults to 2.
            max_ops: Maximum number of operations. Defaults to 5.
            max_distortive: Maximum number of distortive operations. Defaults to 1.
            interpolation: Interpolation method. Defaults to InterpolationMode.NEAREST.
            augmentation_space: Augmentation space from where transforms and magntitudes
                are sampled. Defaults to the one used in the original paper.
            fill: Fill for distortive operations. Defaults to [128, 128, 128].
        """

        self.__p = p
        self.__min_ops = min_ops
        self.__max_ops = max_ops
        self.__max_affine = max_affine
        self.__interpolation = interpolation
        self.__fill = fill
        self.__augmentation_space = augmentation_space
        self.__last_operations = []

    def __call__(self, image: Image.Image) -> Image.Image:
        if not isinstance(image, Image.Image):
            raise TypeError("Expected a PIL image not {}".format(type(image)))
        # Define possible operations.
        operation_pool = list(self.__augmentation_space.keys())
        # Start transforming images.
        num_distortive = 0
        num_operations = 0
        self.__last_operations = []
        while len(operation_pool) > 0:
            # Select random operation and remove from the pool.
            name = random.choice(operation_pool)
            operation_pool.remove(name)
            # Check if operation is allowed.
            if name in AFFINE_TRANSFORMS:
                if num_distortive >= self.__max_affine:
                    # Operation not allowed.
                    continue
                else:
                    num_distortive += 1
            # Get magnitude.
            magnitude = get_magntiude(*self.__augmentation_space[name])
            # Apply transformation.
            image = apply_op(
                image,
                name,
                magnitude,
                interpolation=self.__interpolation,
                fill=self.__fill,
            )
            # Save transform.
            self.__last_operations.append((name, magnitude))
            # Increase operation counter.
            num_operations += 1
            # Check if we continue.
            if num_operations >= self.__min_ops:
                if num_operations >= self.__max_ops:
                    break
                elif random.random() > self.__p:
                    break
        return image

    def repeat(self, image: Image.Image) -> Image.Image:
        if not isinstance(image, Image.Image):
            raise TypeError("Expected a PIL image not {}".format(type(image)))
        for name, magnitude in self.__last_operations:
            image = apply_op(
                image,
                name,
                magnitude,
                interpolation=self.__interpolation,
                fill=self.__fill,
            )
        return image

    def __repr__(self):
        return "{}(p={}, min_ops={}, max_ops={}, max_distortive={})".format(
            self.__class__.__name__,
            self.__p,
            self.__min_ops,
            self.__max_ops,
            self.__max_affine,
        )


def get_magntiude(
    low: Union[float, int, bool], high: Union[float, int, bool]
) -> Union[float, int, bool]:
    if isinstance(low, float):
        return random.uniform(low, high)
    elif isinstance(low, bool):
        return random.choice([True, False])
    elif isinstance(low, int):
        return random.choice(range(low, high + 1))
    else:
        raise ValueError(
            "Augmentation space should contain int/float/bool, not  {}".format(
                type(low)
            )
        )


def augmentation_collage(
    image: Image.Image,
    augmentation: Callable,
    nrows: int = 4,
    ncols: int = 16,
    shape: Tuple[int, int] = (64, 64),
):
    """Generate a collage image with the passed augmentation strategy.

    Args:
        image: Input image.
        auto_augment: AutoAugement strategy
        nrows: Number of collage rows. Defaults to 4.
        ncols: Number of collage columns. Defaults to 16.
        shape: Shape of each image. Defaults to (64, 64).

    Returns:
        Collage image.
    """
    if not isinstance(image, Image.Image):
        raise TypeError("Expected a PIL image not {}".format(type(image)))
    collage = []
    row = []
    for i in range(nrows * ncols):
        transformed = augmentation(image)
        row.append(transformed.resize(shape))
        if len(row) == ncols:
            collage.append(numpy.hstack([numpy.array(x) for x in row]))
            row = []
    collage = numpy.vstack(collage)
    return Image.fromarray(collage)



import numpy
from strong_augment import EXAMPLE_IMAGE, StrongAugment


def test_repeat_agumentation():
    trnsf = StrongAugment()
    for __ in range(128):
        trnsf(EXAMPLE_IMAGE)
    augmented = numpy.array(trnsf(EXAMPLE_IMAGE))
    repeated = numpy.array(trnsf.repeat(EXAMPLE_IMAGE))
    assert numpy.equal(augmented, repeated).all()
    trnsf = StrongAugment(max_affine=0)
    augmented = numpy.array(trnsf(EXAMPLE_IMAGE))
    repeated = numpy.array(trnsf.repeat(EXAMPLE_IMAGE))
    assert numpy.equal(augmented, repeated).all()



import setuptools

__version__ = "0.0.1"

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="strong_augment",
    version=__version__,
    author="jopo666",
    author_email="jopo@birdlover.com",
    description="StrongAugment implementation.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jopo666/StrongAugment",
    packages=setuptools.find_packages(include=["strong_augment", "strong_augment.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="augmentation",
    python_requires=">=3.9",
    install_requires=[
        "torchvision>=0.12",
        "Pillow>=9.1",
    ],
)



import numpy
from strong_augment import EXAMPLE_IMAGE, apply_op
from strong_augment._operations import adjust_channel, gaussian_blur

IMAGE_ARRAY = numpy.array(EXAMPLE_IMAGE)


def test_adjust_channels():
    # magnitude 0.
    assert (
        numpy.array(adjust_channel(EXAMPLE_IMAGE.convert("L"), 0, channel=4235)) == 0
    ).all()
    assert (numpy.array(adjust_channel(EXAMPLE_IMAGE, 0, channel=0)) == 0)[..., 0].all()
    assert (numpy.array(adjust_channel(EXAMPLE_IMAGE, 0, channel=1)) == 0)[..., 1].all()
    assert (numpy.array(adjust_channel(EXAMPLE_IMAGE, 0, channel=2)) == 0)[..., 2].all()
    # no operation.
    assert (
        numpy.array(adjust_channel(EXAMPLE_IMAGE, 1.0, channel=0)) == IMAGE_ARRAY
    ).all()
    assert (
        numpy.array(adjust_channel(EXAMPLE_IMAGE, 1.0, channel=1)) == IMAGE_ARRAY
    ).all()
    assert (
        numpy.array(adjust_channel(EXAMPLE_IMAGE, 1.0, channel=2)) == IMAGE_ARRAY
    ).all()
    # halve.
    assert (
        numpy.array(adjust_channel(EXAMPLE_IMAGE, 0.5, channel=2))[..., 2]
        == (IMAGE_ARRAY[..., 2] * 0.5).astype(numpy.uint8)
    ).all()


def test_gaussian_blur_no_operation():
    assert (
        numpy.array(gaussian_blur(EXAMPLE_IMAGE)) == numpy.array(EXAMPLE_IMAGE)
    ).all()


def test_brightness_contrast_gamma_zero():
    assert numpy.array(apply_op(EXAMPLE_IMAGE, "brightness", 0)).mean() == 0
    assert numpy.array(apply_op(EXAMPLE_IMAGE, "contrast", 0)).mean() == 128
    assert numpy.array(apply_op(EXAMPLE_IMAGE, "gamma", 0)).mean() == 255


def test_no_op():
    assert (numpy.array(apply_op(EXAMPLE_IMAGE, "red", 1.0)) == IMAGE_ARRAY).all()
    assert (numpy.array(apply_op(EXAMPLE_IMAGE, "green", 1.0)) == IMAGE_ARRAY).all()
    assert (numpy.array(apply_op(EXAMPLE_IMAGE, "blue", 1.0)) == IMAGE_ARRAY).all()
    assert (numpy.array(apply_op(EXAMPLE_IMAGE, "hue", 0.0)) == IMAGE_ARRAY).all()
    assert (numpy.array(apply_op(EXAMPLE_IMAGE, "gamma", 1.0)) == IMAGE_ARRAY).all()
    assert (numpy.array(apply_op(EXAMPLE_IMAGE, "contrast", 1.0)) == IMAGE_ARRAY).all()
    assert (numpy.array(apply_op(EXAMPLE_IMAGE, "sharpness", 1)) == IMAGE_ARRAY).all()
    assert (numpy.array(apply_op(EXAMPLE_IMAGE, "blur", 0)) == IMAGE_ARRAY).all()
    assert (numpy.array(apply_op(EXAMPLE_IMAGE, "posterize", 8)) == IMAGE_ARRAY).all()
    assert (numpy.array(apply_op(EXAMPLE_IMAGE, "solarize", 256)) == IMAGE_ARRAY).all()
    assert (
        numpy.array(apply_op(EXAMPLE_IMAGE, "saturation", 1.0)) == IMAGE_ARRAY
    ).all()
    assert (
        numpy.array(apply_op(EXAMPLE_IMAGE, "brightness", 1.0)) == IMAGE_ARRAY
    ).all()
